/*
 * Copyright (c) 2015, 2016, 2017 Adrian Siekierka
 *
 * This file is part of Preston.
 *
 * Preston is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Preston is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Preston.  If not, see <http://www.gnu.org/licenses/>.
 */

package pl.asie.preston.network;

import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;
import net.minecraft.entity.Entity;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.entity.player.EntityPlayerMP;
import net.minecraft.server.management.PlayerChunkMap;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;
import net.minecraft.world.WorldServer;
import net.minecraftforge.fml.common.network.FMLEmbeddedChannel;
import net.minecraftforge.fml.common.network.FMLOutboundHandler;
import net.minecraftforge.fml.common.network.NetworkRegistry;
import net.minecraftforge.fml.common.network.NetworkRegistry.TargetPoint;
import net.minecraftforge.fml.relauncher.Side;
import pl.asie.preston.PrestonMod;

import java.util.EnumMap;
import java.util.HashSet;

public class PacketRegistry {
	private static final HashSet<String> usedChannelNames = new HashSet<>();

	private EnumMap<Side, FMLEmbeddedChannel> channels;
	@SuppressWarnings("unchecked")
	private Class<? extends Packet>[] idPacketMap = (Class<? extends Packet>[]) new Class[256];
	private Object2IntMap<Class<? extends Packet>> packetIdMap = new Object2IntOpenHashMap<>();

	public PacketRegistry(String channelName) {
		if (channelName.length() > 20) {
			String trunc = channelName.substring(0, 20);
			PrestonMod.logger.warn("Channel name too long: " + channelName + ", truncating to " + trunc);
			channelName = trunc;
		}
		if (usedChannelNames.contains(channelName)) {
			throw new RuntimeException("Channel name already used: " + channelName);
		}

		channels = NetworkRegistry.INSTANCE.newChannel(channelName, new PacketChannelHandler(this));
		usedChannelNames.add(channelName);
	}

	public void registerPacket(int id, Class<? extends Packet> packet) {
		try {
			packet.getConstructor();
		} catch (NoSuchMethodException e) {
			throw new RuntimeException("No empty constructor defined! This is a Charset bug!", e);
		}
		idPacketMap[id] = packet;
		packetIdMap.put(packet, id);
	}

	public net.minecraft.network.Packet getPacketFrom(Packet message) {
		return channels.get(Side.SERVER).generatePacketFrom(message);
	}

	public void sendToAll(Packet message) {
		channels.get(Side.SERVER).attr(FMLOutboundHandler.FML_MESSAGETARGET).set(FMLOutboundHandler.OutboundTarget.ALL);
		channels.get(Side.SERVER).writeOutbound(message);
	}

	public void sendToWatching(Packet message, World world, BlockPos pos, Entity except) {
		WorldServer worldServer = (WorldServer) world;
		PlayerChunkMap map = worldServer.func_184164_w();
		for (EntityPlayer player : worldServer.field_73010_i) {
			if (player != except) {
				if (map.func_72694_a((EntityPlayerMP) player, pos.func_177958_n() >> 4, pos.func_177952_p() >> 4)) {
					channels.get(Side.SERVER).attr(FMLOutboundHandler.FML_MESSAGETARGET).set(FMLOutboundHandler.OutboundTarget.PLAYER);
					channels.get(Side.SERVER).attr(FMLOutboundHandler.FML_MESSAGETARGETARGS).set(player);
					channels.get(Side.SERVER).writeOutbound(message);
				}
			}
		}
	}

	public void sendToWatching(Packet message, TileEntity tile) {
		sendToWatching(message, tile.func_145831_w(), tile.func_174877_v(), null);
	}

	public void sendToWatching(Packet message, Entity entity) {
		sendToWatching(message, entity.func_130014_f_(), entity.func_180425_c(), null);
	}

	// TODO: Slightly ugly.
	public void sendToWatching(Packet message, TileEntity tile, Entity except) {
		sendToWatching(message, tile.func_145831_w(), tile.func_174877_v(), except);
	}

	public void sendToWatching(Packet message, Entity entity, Entity except) {
		sendToWatching(message, entity.func_130014_f_(), entity.func_180425_c(), except);
	}

	public void sendTo(Packet message, EntityPlayer player) {
		channels.get(Side.SERVER).attr(FMLOutboundHandler.FML_MESSAGETARGET).set(FMLOutboundHandler.OutboundTarget.PLAYER);
		channels.get(Side.SERVER).attr(FMLOutboundHandler.FML_MESSAGETARGETARGS).set(player);
		channels.get(Side.SERVER).writeOutbound(message);
	}

	public void sendToAllAround(Packet message, NetworkRegistry.TargetPoint point) {
		channels.get(Side.SERVER).attr(FMLOutboundHandler.FML_MESSAGETARGET).set(FMLOutboundHandler.OutboundTarget.ALLAROUNDPOINT);
		channels.get(Side.SERVER).attr(FMLOutboundHandler.FML_MESSAGETARGETARGS).set(point);
		channels.get(Side.SERVER).writeOutbound(message);
	}

	public void sendToDimension(Packet message, int dimensionId) {
		channels.get(Side.SERVER).attr(FMLOutboundHandler.FML_MESSAGETARGET).set(FMLOutboundHandler.OutboundTarget.DIMENSION);
		channels.get(Side.SERVER).attr(FMLOutboundHandler.FML_MESSAGETARGETARGS).set(dimensionId);
		channels.get(Side.SERVER).writeOutbound(message);
	}

	public void sendToServer(Packet message) {
		channels.get(Side.CLIENT).attr(FMLOutboundHandler.FML_MESSAGETARGET).set(FMLOutboundHandler.OutboundTarget.TOSERVER);
		channels.get(Side.CLIENT).writeOutbound(message);
	}

	public void sendToAllAround(Packet packet, TileEntity entity,
								double d) {
		this.sendToAllAround(packet, new TargetPoint(entity.func_145831_w().field_73011_w.getDimension(),
				entity.func_174877_v().func_177958_n(), entity.func_174877_v().func_177956_o(), entity.func_174877_v().func_177952_p(), d));
	}

	public void sendToAllAround(Packet packet, Entity entity,
								double d) {
		this.sendToAllAround(packet, new TargetPoint(entity.field_71093_bK, entity.field_70165_t, entity.field_70163_u, entity.field_70161_v, d));
	}

	public Packet instantiatePacket(int i) {
		try {
			return idPacketMap[i].newInstance();
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public int getPacketId(Class<? extends Packet> aClass) {
		return packetIdMap.get(aClass);
	}
}
